use async_trait::async_trait;
use nostr::{Event, JsonUtil};
use sea_query::{extension::sqlite::SqliteExpr, Query};
use sea_query_binder::SqlxBinder;
use sqlx::sqlite::{Sqlite, SqlitePoolOptions};
use sqlx::FromRow;
use sqlx::{migrate::MigrateDatabase, Pool};
use std::sync::Arc;

use super::db::Noose;
use crate::bussy::{channels, Command, Message, PubSub};
use crate::utils::{error::Error, structs::Subscription};

enum EventsTable {
    Table,
    EventId,
    Kind,
    Pubkey,
    Content,
    CreatedAt,
    Tags,
    Sig,
}

impl sea_query::Iden for EventsTable {
    fn unquoted(&self, s: &mut dyn std::fmt::Write) {
        write!(
            s,
            "{}",
            match self {
                Self::Table => "events",
                Self::EventId => "id",
                Self::Kind => "kind",
                Self::Pubkey => "pubkey",
                Self::Content => "content",
                Self::CreatedAt => "created_at",
                Self::Tags => "tags",
                Self::Sig => "sig",
            }
        )
        .unwrap()
    }
}

enum EventsFTSTable {
    Table,
    EventId,
    Content,
}

impl sea_query::Iden for EventsFTSTable {
    fn unquoted(&self, s: &mut dyn std::fmt::Write) {
        write!(
            s,
            "{}",
            match self {
                Self::Table => "events_fts",
                Self::EventId => "id",
                Self::Content => "content",
            }
        )
        .unwrap()
    }
}

enum TagsTable {
    Table,
    Tag,
    Value,
    EventId,
}

impl sea_query::Iden for TagsTable {
    fn unquoted(&self, s: &mut dyn std::fmt::Write) {
        write!(
            s,
            "{}",
            match self {
                Self::Table => "tags",
                Self::Tag => "tag",
                Self::Value => "value",
                Self::EventId => "event_id",
            }
        )
        .unwrap()
    }
}

#[derive(FromRow, Debug)]
struct EventsCountRow(i32);

#[derive(FromRow, Debug)]
struct EventRow {
    id: String,
    pubkey: String,
    created_at: i64,
    kind: i64,
    tags: String,
    sig: String,
    content: String,
}

impl EventRow {
    pub fn to_string(&self, subscription_id: nostr::SubscriptionId) -> String {
        let tags: Vec<Vec<String>> = serde_json::from_str(&self.tags).unwrap();

        let message = serde_json::json!([
            "EVENT",
            subscription_id,
            {
                "id": self.id,
                "content": self.content,
                "created_at": self.created_at,
                "kind": self.kind,
                "pubkey": self.pubkey,
                "sig": self.sig,
                "tags": tags
            }
        ]);

        message.to_string()
    }
}

pub struct SqliteDb {
    pool: Pool<Sqlite>,
}

impl SqliteDb {
    pub async fn new() -> Self {
        let pool = SqliteDb::build_pool("noose_pool", 42).await;

        Self { pool }
    }

    pub fn info(&self) {
        dbg!(self.pool.options());
    }

    async fn migrate(pool: &Pool<Sqlite>) {
        sqlx::migrate!("src/noose/migrations")
            .run(pool)
            .await
            .unwrap()
    }

    async fn build_pool(name: &str, max_size: u32) -> Pool<Sqlite> {
        let pool_options = SqlitePoolOptions::new()
            .test_before_acquire(true)
            // .idle_timeout(Some(Duration::from_secs(10)))
            // .max_lifetime(Some(Duration::from_secs(30)))
            .max_lifetime(None)
            .idle_timeout(None)
            .max_connections(max_size);

        let db_url = "sqlite://sqlite.db";
        if !Sqlite::database_exists(db_url).await.unwrap_or(false) {
            log::info!("Creating database {}", db_url);
            match Sqlite::create_database(db_url).await {
                Ok(_) => log::info!("Db {} created", db_url),
                Err(_) => panic!("Failed to create database {}", db_url),
            }
        }
        if let Ok(pool) = pool_options.connect(db_url).await {
            log::info!("Connected to sqlite pool {}", name);

            pool
        } else {
            panic!("Connection to sqlite pool {} failed", name);
        }
    }

    async fn add_event(&self, event: Box<Event>) -> Result<String, Error> {
        let id = event.id.to_string();
        let kind = event.kind.to_string();
        let pubkey = event.pubkey.to_string();
        let content = event.content.to_string();
        let created_at = event.created_at.as_i64();
        let tags = serde_json::to_string(&event.tags).unwrap();
        let sig = event.sig.to_string();

        let message = format!("[\"OK\", \"{}\", true, \"\"]", id.clone());

        if event.is_ephemeral() {
            return Ok(message);
        }

        let tx = self.pool.begin().await.unwrap();
        {
            if event.is_replaceable() {
                dbg!("new event is replaceable - searching for previously stored event");
                let (sql, values) = Query::select()
                    .from(EventsTable::Table)
                    .columns([EventsTable::EventId])
                    .and_where(
                        sea_query::Expr::col(EventsTable::Pubkey).eq(event.pubkey.to_string()),
                    )
                    .and_where(sea_query::Expr::col(EventsTable::Kind).eq(event.kind.as_u32()))
                    .and_where(
                        sea_query::Expr::col(EventsTable::CreatedAt).gte(event.created_at.as_i64()),
                    )
                    .limit(1)
                    .build_sqlx(sea_query::SqliteQueryBuilder);

                let repl_count = sqlx::query_with(&sql, values).fetch_one(&self.pool).await;

                if repl_count.ok().is_some() {
                    return Ok(message);
                }
            }
        }

        {
            if event.is_parameterized_replaceable() {
                dbg!(
                    "new event is parametrized replaceable - searching for previously stored event"
                );
                let d_tags: Vec<String> = event
                    .tags
                    .iter()
                    .filter(|tag| tag.kind() == nostr::TagKind::D)
                    .map(|tag| tag.clone().to_vec()[1].clone())
                    .collect();
                let (sql, values) = Query::select()
                    .from(EventsTable::Table)
                    .column((EventsTable::Table, EventsTable::EventId))
                    .left_join(
                        TagsTable::Table,
                        sea_query::Expr::col((TagsTable::Table, TagsTable::EventId))
                            .equals((EventsTable::Table, EventsTable::EventId)),
                    )
                    .and_where(
                        sea_query::Expr::col((EventsTable::Table, EventsTable::Pubkey))
                            .eq(event.pubkey.to_string()),
                    )
                    .and_where(
                        sea_query::Expr::col((EventsTable::Table, EventsTable::Kind))
                            .eq(event.kind.as_u32()),
                    )
                    .and_where(sea_query::Expr::col((TagsTable::Table, TagsTable::Tag)).eq("d"))
                    .and_where(
                        sea_query::Expr::col((TagsTable::Table, TagsTable::Value))
                            .eq(d_tags[0].to_string()),
                    )
                    .and_where(
                        sea_query::Expr::col((EventsTable::Table, EventsTable::CreatedAt))
                            .gte(event.created_at.as_i64()),
                    )
                    .limit(1)
                    .build_sqlx(sea_query::SqliteQueryBuilder);

                let repl_count = sqlx::query_with(&sql, values).fetch_one(&self.pool).await;

                if repl_count.ok().is_some() {
                    return Ok(message);
                }
            }
        }

        // Insert replaceble event
        {
            if event.is_replaceable() {
                dbg!("deleting older replaceable event from events table");
                let (sql, values) = Query::delete()
                    .from_table(EventsTable::Table)
                    .and_where(
                        sea_query::Expr::col((EventsTable::Table, EventsTable::Kind))
                            .eq(event.kind.as_u32()),
                    )
                    .and_where(
                        sea_query::Expr::col((EventsTable::Table, EventsTable::Pubkey))
                            .eq(event.pubkey.to_string()),
                    )
                    .and_where(
                        sea_query::Expr::col((EventsTable::Table, EventsTable::EventId))
                            .not_in_subquery(
                                Query::select()
                                    .from(EventsTable::Table)
                                    .column(EventsTable::EventId)
                                    .and_where(
                                        sea_query::Expr::col(EventsTable::Kind)
                                            .eq(event.kind.as_u32()),
                                    )
                                    .and_where(
                                        sea_query::Expr::col(EventsTable::Pubkey)
                                            .eq(event.pubkey.to_string()),
                                    )
                                    .order_by(EventsTable::CreatedAt, sea_query::Order::Desc)
                                    .limit(1)
                                    .to_owned(),
                            ),
                    )
                    .build_sqlx(sea_query::SqliteQueryBuilder);

                let results = sqlx::query_with(&sql, values)
                    .execute(&self.pool)
                    .await
                    .unwrap();

                if results.rows_affected() > 0 {
                    log::info!(
                        "removed {} older replaceable kind {} events for author: {:?}",
                        results.rows_affected(),
                        event.kind.as_u32(),
                        event.pubkey.to_string()
                    );
                }
            }
        }

        // Insert parametrized replaceble event
        {
            if event.is_parameterized_replaceable() {
                dbg!("deleting older parametrized replaceable event from events table");
                let d_tag = event.identifier();
                let (sql, values) = Query::delete()
                    .from_table(EventsTable::Table)
                    .and_where(
                        sea_query::Expr::col((EventsTable::Table, EventsTable::EventId))
                            .in_subquery(
                                Query::select()
                                    .from(EventsTable::Table)
                                    .column((EventsTable::Table, EventsTable::EventId))
                                    .left_join(
                                        TagsTable::Table,
                                        sea_query::Expr::col((
                                            TagsTable::Table,
                                            TagsTable::EventId,
                                        ))
                                        .equals((EventsTable::Table, EventsTable::EventId)),
                                    )
                                    .and_where(
                                        sea_query::Expr::col((
                                            EventsTable::Table,
                                            EventsTable::Kind,
                                        ))
                                        .eq(event.kind.as_u32()),
                                    )
                                    .and_where(
                                        sea_query::Expr::col((
                                            EventsTable::Table,
                                            EventsTable::Pubkey,
                                        ))
                                        .eq(event.pubkey.to_string()),
                                    )
                                    .and_where(
                                        sea_query::Expr::col((TagsTable::Table, TagsTable::Tag))
                                            .eq("d"),
                                    )
                                    .and_where(
                                        sea_query::Expr::col((TagsTable::Table, TagsTable::Value))
                                            .eq(d_tag),
                                    )
                                    .to_owned(),
                            ),
                    )
                    .build_sqlx(sea_query::SqliteQueryBuilder);

                let results = sqlx::query_with(&sql, values)
                    .execute(&self.pool)
                    .await
                    .unwrap();

                if results.rows_affected() > 0 {
                    log::info!("removed {} older parameterized replaceable kind {} events for author: {:?}", results.rows_affected(), event.kind, event.pubkey);
                }
            }
        }

        // Process deletion events
        dbg!(event.as_json());
        if event.kind.as_u32() == 5 {
            dbg!("deleting event");
            let ids: Vec<String> = event.event_ids().map(|eid| eid.to_string()).collect();
            let (sql, values) = Query::delete()
                .from_table(EventsTable::Table)
                .and_where(sea_query::Expr::col(EventsTable::Kind).ne(5))
                .and_where(sea_query::Expr::col(EventsTable::Pubkey).eq(event.pubkey.to_string()))
                .and_where(sea_query::Expr::col(EventsTable::EventId).is_in(&ids))
                .build_sqlx(sea_query::SqliteQueryBuilder);

            let results = sqlx::query_with(&sql, values)
                .execute(&self.pool)
                .await
                .unwrap();

            if results.rows_affected() > 0 {
                log::info!(
                    "removed {} events for author {:?}",
                    results.rows_affected(),
                    event.pubkey
                );
            }

            // Delete from EventsFTS
            let (sql, values) = Query::delete()
                .from_table(EventsFTSTable::Table)
                .and_where(sea_query::Expr::col(EventsFTSTable::EventId).is_in(&ids))
                .build_sqlx(sea_query::SqliteQueryBuilder);
            let _ = sqlx::query_with(&sql, values)
                .execute(&self.pool)
                .await
                .unwrap();
        } else {
            dbg!("inserting new event in events");
            // Insert into Events table
            let (sql, values) = Query::insert()
                .into_table(EventsTable::Table)
                .columns([
                    EventsTable::EventId,
                    EventsTable::Content,
                    EventsTable::Kind,
                    EventsTable::Pubkey,
                    EventsTable::CreatedAt,
                    EventsTable::Tags,
                    EventsTable::Sig,
                ])
                .values_panic([
                    id.clone().into(),
                    content.clone().into(),
                    kind.into(),
                    pubkey.into(),
                    created_at.into(),
                    tags.into(),
                    sig.into(),
                ])
                .build_sqlx(sea_query::SqliteQueryBuilder);

            let results = sqlx::query_with(&sql, values)
                .execute(&self.pool)
                .await
                .unwrap();

            // Insert into EventsFTS table
            dbg!("inserting new event into eventsFTS");
            let (sql, values) = Query::insert()
                .into_table(EventsFTSTable::Table)
                .columns([EventsFTSTable::EventId, EventsFTSTable::Content])
                .values_panic([id.clone().into(), content.into()])
                .build_sqlx(sea_query::SqliteQueryBuilder);

            let results = sqlx::query_with(&sql, values)
                .execute(&self.pool)
                .await
                .unwrap();

            // Insert into Tags table
            dbg!("inserting new event into tags");
            for tag in event.tags.clone() {
                let tag = tag.to_vec();
                if tag.len() >= 2 {
                    let tag_name = &tag[0];
                    let tag_value = &tag[1];
                    if tag_name.len() == 1 {
                        let (sql, values) = Query::insert()
                            .into_table(TagsTable::Table)
                            .columns([TagsTable::Tag, TagsTable::Value, TagsTable::EventId])
                            .values_panic([tag_name.into(), tag_value.into(), id.clone().into()])
                            .build_sqlx(sea_query::SqliteQueryBuilder);

                        let results = sqlx::query_with(&sql, values)
                            .execute(&self.pool)
                            .await
                            .unwrap();
                    }
                }
            }
        }

        tx.commit().await.unwrap();

        Ok(message)
    }

    async fn index_search(&self, event: Box<Event>) -> Result<(), Error> {
        let id = event.id.to_string();
        let content = event.content.to_string();
        let (sql, values) = Query::insert()
            .into_table(EventsFTSTable::Table)
            .columns([EventsFTSTable::EventId, EventsFTSTable::Content])
            .values_panic([id.into(), content.into()])
            .build_sqlx(sea_query::SqliteQueryBuilder);

        let results = sqlx::query_with(&sql, values).execute(&self.pool).await;
        if results.is_ok() {
            Ok(())
        } else {
            Err(Error::internal_with_message(
                "Unable to write event to events_fts index",
            ))
        }
    }

    async fn index_tags(&self, event: Box<Event>) -> Result<(), Error> {
        // let t: Vec<String> = Vec::new();
        // for tag in event.tags {
        //     tag.kind()
        // }

        Ok(())
    }

    fn get_filter_query(&self, filter: &nostr::Filter) -> sea_query::SelectStatement {
        let mut query = Query::select()
            .column((EventsTable::Table, EventsTable::EventId))
            .column((EventsTable::Table, EventsTable::Content))
            .columns([
                EventsTable::Kind,
                EventsTable::Pubkey,
                EventsTable::CreatedAt,
                EventsTable::Tags,
                EventsTable::Sig,
            ])
            .from(EventsTable::Table)
            .order_by(EventsTable::CreatedAt, sea_query::Order::Desc)
            .to_owned();

        if !filter.ids.is_empty() {
            let ids = filter.ids.iter().map(|id| id.to_string());
            query = query
                .and_where(
                    sea_query::Expr::col((EventsTable::Table, EventsTable::EventId)).is_in(ids),
                )
                .to_owned();
        }

        if !filter.kinds.is_empty() {
            let kinds: Vec<u32> = filter.kinds.iter().map(|kind| kind.as_u32()).collect();
            query = query
                .and_where(
                    sea_query::Expr::col((EventsTable::Table, EventsTable::Kind)).is_in(kinds),
                )
                .to_owned();
        }

        if !filter.authors.is_empty() {
            let authors = filter.authors.iter().map(|author| author.to_string());
            query = query
                .and_where(
                    sea_query::Expr::col((EventsTable::Table, EventsTable::Pubkey)).is_in(authors),
                )
                .to_owned();
        }

        if let Some(since) = filter.since {
            query = query
                .and_where(
                    sea_query::Expr::col((EventsTable::Table, EventsTable::CreatedAt))
                        .gte(since.as_u64()),
                )
                .to_owned();
        }

        if let Some(until) = filter.until {
            query = query
                .and_where(
                    sea_query::Expr::col((EventsTable::Table, EventsTable::CreatedAt))
                        .lte(until.as_u64()),
                )
                .to_owned();
        }

        if let Some(limit) = filter.limit {
            query = query.limit(limit as u64).to_owned();
        }

        filter.generic_tags.iter().for_each(|(tag, values)| {
            let values = values.iter().map(|val| val.to_string());
            query = query
                .left_join(
                    TagsTable::Table,
                    sea_query::Expr::col((TagsTable::Table, TagsTable::EventId))
                        .equals((EventsTable::Table, EventsTable::EventId)),
                )
                .and_where(
                    sea_query::Expr::col((TagsTable::Table, TagsTable::Tag)).eq(tag.to_string()),
                )
                .and_where(sea_query::Expr::col((TagsTable::Table, TagsTable::Value)).is_in(values))
                .to_owned();
        });

        if let Some(search) = &filter.search {
            query = query
                .inner_join(
                    EventsFTSTable::Table,
                    sea_query::Expr::col((EventsTable::Table, EventsTable::EventId))
                        .equals((EventsFTSTable::Table, EventsFTSTable::EventId)),
                )
                .and_where(
                    sea_query::Expr::col((EventsFTSTable::Table, EventsFTSTable::Content))
                        .matches(search),
                )
                .to_owned();
        }

        query
    }

    fn get_filters_query(&self, subscription: Subscription) -> Option<sea_query::SelectStatement> {
        subscription
            .filters
            .iter()
            .map(|filter| {
                Query::select()
                    .column((EventsTable::Table, EventsTable::EventId))
                    .column((EventsTable::Table, EventsTable::Content))
                    .columns([
                        EventsTable::Kind,
                        EventsTable::Pubkey,
                        EventsTable::CreatedAt,
                        EventsTable::Tags,
                        EventsTable::Sig,
                    ])
                    .from_subquery(
                        self.get_filter_query(filter),
                        sea_query::Alias::new("events"),
                    )
                    .to_owned()
            })
            .reduce(|mut result, query| result.union(sea_query::UnionType::All, query).to_owned())
    }

    async fn delete_filters(&self, subscription: Subscription) -> Vec<EventRow> {
        todo!()
    }

    async fn count_events_by_filters(&self, subscription: Subscription) -> i32 {
        if subscription.filters.is_empty() {
            return 0;
        }

        let (sql, values) = self
            .get_filters_query(subscription)
            .unwrap()
            .clear_selects()
            .expr_as(
                sea_query::Func::count(sea_query::Expr::col((
                    EventsTable::Table,
                    EventsTable::EventId,
                ))),
                sea_query::Alias::new("count"),
            )
            .build_sqlx(sea_query::SqliteQueryBuilder);

        println!("count_filters SEA_QUERY built SQL: {}", sql.clone());

        let counts = sqlx::query_as_with::<_, EventsCountRow, _>(&sql, values)
            .fetch_one(&self.pool)
            .await
            .unwrap();

        dbg!(counts);

        1
    }
}

#[async_trait]
impl Noose for SqliteDb {
    async fn start(&mut self, pubsub: Arc<PubSub>) -> Result<(), Error> {
        let mut subscriber = pubsub.subscribe(channels::MSG_NOOSE).await;

        while let Ok(message) = subscriber.recv().await {
            log::info!("[Noose] received message: {:?}", message);
            let command = match message.content {
                Command::DbReqWriteEvent(event) => match self.write_event(event).await {
                    Ok(status) => Command::DbResOkWithStatus(status),
                    Err(e) => Command::ServiceError(e),
                },
                Command::DbReqFindEvent(client_id, subscriptioin) => {
                    match self.find_event(subscriptioin).await {
                        Ok(events) => Command::DbResRelayMessage(client_id, events),
                        Err(e) => Command::ServiceError(e),
                    }
                }
                _ => Command::Noop,
            };
            if command != Command::Noop {
                let channel = message.source;
                let message = Message {
                    source: channels::MSG_NOOSE,
                    content: command,
                };

                log::debug!("[Noose] publishing new message: {:?}", message);

                pubsub.publish(channel, message).await;
            }
        }

        Ok(())
    }

    async fn migration_up(&self) {
        SqliteDb::migrate(&self.pool).await;
    }

    async fn write_event(&self, event: Box<Event>) -> Result<String, Error> {
        log::debug!("[Noose] write_event triggered");

        let status = self.add_event(event).await.unwrap();

        return Ok(status);
    }

    async fn find_event(&self, subscription: Subscription) -> Result<Vec<String>, Error> {
        log::debug!("making query from filters...");

        let eose_message =
            vec![nostr::RelayMessage::EndOfStoredEvents(subscription.id.clone()).as_json()];

        if let Some(sql_statement) = self.get_filters_query(subscription.clone()) {
            let (sql, values) = sql_statement.build_sqlx(sea_query::SqliteQueryBuilder);
            log::info!("SEA_QUERY built SQL: {}", sql.clone());

            match sqlx::query_as_with::<_, EventRow, _>(&sql, values)
                .fetch_all(&self.pool)
                .await
            {
                Ok(rows) => {
                    if rows.is_empty() {
                        return Ok(eose_message);
                    } else {
                        let relay_messages: Vec<String> = rows
                            .iter()
                            .map(|row| row.to_string(subscription.id.clone()))
                            .collect();
                        return Ok(relay_messages);
                    }
                }
                Err(e) => {
                    log::error!("{}", e);
                    return Err(Error::internal(e.into()));
                }
            }
        }

        return Ok(eose_message);
    }
}

#[cfg(test)]
mod tests {
    use super::Noose;
    use super::SqliteDb;
    use crate::utils::structs::Subscription;

    use nostr::util::JsonUtil;

    #[tokio::test]
    async fn find_event() {
        let db = SqliteDb::new().await;

        let t = std::time::Instant::now();
        let client_id = "test_id".to_string();

        let cm = nostr::ClientMessage::from_json(
            r#"["REQ","7b9bc4b6-701c-40b6-898f-4e7c6b5b1510",{"authors":["04c915daefee38317fa734444acee390a8269fe5810b2241e5e6dd343dfbecc9"],"kinds":[0]}]"#,
        ).unwrap();
        let (sub_id, filters) = match cm {
            nostr::ClientMessage::Req {
                subscription_id,
                filters,
            } => (subscription_id, filters),
            _ => panic!("sneed :("),
        };
        let sub = Subscription::new(sub_id, filters);
        db.find_event(sub).await.unwrap();
        println!(
            "Time passed: {}",
            (std::time::Instant::now() - t).as_millis()
        );
    }

    #[tokio::test]
    async fn delete_events() {
        let db = SqliteDb::new().await;

        let t = std::time::Instant::now();
        let client_id = "test_id".to_string();

        let my_keys = nostr::Keys::generate();

        let eid = nostr::EventId::all_zeros();
        let tag_event = nostr::Tag::Event {
            event_id: eid,
            relay_url: None,
            marker: None,
        };
        let tag_url = nostr::Tag::AbsoluteURL(nostr::types::UncheckedUrl::new(
            "http://foo.net".to_string(),
        ));
        let tag_hashtag = nostr::Tag::Hashtag("farm".to_string());

        let event = nostr::EventBuilder::new_text_note(
            "sneed feed and seed",
            vec![tag_event, tag_url, tag_hashtag],
        )
        .to_event(&my_keys)
        .unwrap();

        dbg!(&event.as_json());
        let resp = db.add_event(Box::new(event.clone())).await.unwrap();
        dbg!(resp);

        let delete_event = nostr::EventBuilder::delete(vec![event.id])
            .to_event(&my_keys)
            .unwrap();

        dbg!(&delete_event);
        let resp = db.add_event(Box::new(delete_event.clone())).await.unwrap();
        dbg!(resp);

        // let sub_id = nostr::SubscriptionId::new("test".to_string());
        // let mut subscription = Subscription::new(sub_id, vec![]);

        // if delete_event.kind == nostr::Kind::EventDeletion {
        //     delete_event
        //         .tags
        //         .iter()
        //         .filter(|tag| {
        //             matches!(
        //                 tag,
        //                 nostr::Tag::Event {
        //                     event_id,
        //                     relay_url,
        //                     marker,
        //                 }
        //             )
        //         })
        //         .for_each(|tag| {
        //             if let nostr::Tag::Event {
        //                 event_id,
        //                 relay_url,
        //                 marker,
        //             } = tag
        //             {
        //                 let filter = nostr::Filter::new();
        //                 let filter = &filter.event(*event_id);

        //                 subscription.filters.push(filter.clone());
        //             }
        //         });

        //     dbg!(&subscription);
        // }

        // let res = db.delete_filters(subscription).await;

        // dbg!(res);

        // let sub = Subscription::new(sub_id, filters);
        // let num = db.delete_filters(sub).await.len();
        // println!(
        //     "Time passed: {}",
        //     (std::time::Instant::now() - t).as_millis()
        // );

        // assert_eq!(num, 1);
    }

    #[tokio::test]
    async fn count_events() {
        let db = SqliteDb::new().await;

        let t = std::time::Instant::now();
        let client_id = "test_id".to_string();

        let cm = nostr::ClientMessage::from_json(
            r#"["COUNT","7b9bc4b6-701c-40b6-898f-4e7c6b5b1510",{"authors":["6be3c1446231fe6d117d72e29b60094bbb3eec029100c34f627dc4ebe8369a64"],"kinds":[1]}]"#,
        ).unwrap();
        let (sub_id, filters) = match cm {
            nostr::ClientMessage::Count {
                subscription_id,
                filters,
            } => (subscription_id, filters),
            _ => panic!("sneed :("),
        };

        let sub = Subscription::new(sub_id, filters);
        let num = db.count_events_by_filters(sub).await;
        println!(
            "Time passed: {}",
            (std::time::Instant::now() - t).as_millis()
        );

        assert_eq!(num, 1);
    }
}
